	$TITLE	(Line Time Clock Software Driver)


; Copyright (C) 1994 by Dan Liddell and Robert Armstrong.
; Copyright (C) 2004 by Spare Time Gizmos.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-
; ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
; Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, visit the website of the Free
; Software Foundation, Inc., www.gnu.org.

;   The hardware derives a 120Hz (not 60 Hz!!) clock from the AC power
; line.  This clock is coupled to the INT 1 interrupt of the 8051,
; where it causes an interrupt every 8.3 milliseconds.  This clock
; (also called the "line time clock" to avoid confusion) serves as the
; basis for all the internal timing in the Life game.  We also compute
; the time of day by counting down the line time clock interrupts.


; CLOCK_INTERRUPT
;
;   This routine is called from the interrupt 1 vector in low ROM.
; We have three major things to do: (1) count down the buzzer timeout,
; (2) count down the WAIT routine timeout, and (3) count up the ticks
; for the once-a-second code.  Since this code gets called frequently,
; we'd like to avoid spending much time here.  Things are carefully
; arranged so that the most common cases (when nothing much special
; happens) exit quickly.  The best case takes 24 machine cycles, counting
; the interrupt overhead.
;
;   WARNING!  Remember that this routine, and everything that it calls,
; is an interrupt routine.  All registers and flags that we change
; must be saved and restored!
;
CLOCK_INTERRUPT:
	PUSH	ACC		; [2] we need one register to work with
	PUSH	PSW		; we need to save the carry too
	MOV	A,WAIT_TIME	; [2] get the current WAIT timer
	JZ	CLKIN1		; [2] don't decrement if it's zero
	DEC	WAIT_TIME	; [2] otherwise count it down
CLKIN1:	MOV	A,BUZZER_TIME	; [2] next check the buzzer timeout
	JZ	CLKIN2		; [2] skip out if it's already zero
	DEC	BUZZER_TIME	; [2] otherwise count down the timeout
	CJNE	A, #1, CLKIN2	; [2] did it just become zero ??
	BUZZER_OFF		; [2] yes - turn the buzzer off now!
CLKIN2:	INC	TICKS		; [2] finally, count the clock ticks
	MOV	A, TICKS	; [2] let's see what we've got
	CJNE	A,#HERTZ,CLKIN3	; [2] skip if we haven't done a full second
	ACALL	ONCE_A_SECOND	; [2] otherwise to the once-a-second stuff
CLKIN3: POP	PSW		; get the carry bit back
	POP	ACC		; [2] restore the registers
	CLR	IE1		; just in case
	RETI			; [2] and dismiss the interrupt


; ONCE_A_SECOND
;
;   This routine is called once every second (as if you couldn't tell
; from the name) and keeps track of the time of day.  Currently we
; always keep the time in 24 hour format and we don't worry about the
; date, so the clock resets once every night at midnight.
;
;   REMEMBER that this is an interrupt routine.  At this point, the
; accumulator has been saved and is available for use.  Nothing else
; is safe !!
;
ONCE_A_SECOND:
	MOV	TICKS, #0	; reset the tick counter
	INC	SECONDS		; and count another second gone by
	mov a,ELAPSEDTIME	; will we max out elapsed time with increment?
	inc a
	jz OAS1			; yep. it wrapped around
	inc ELAPSEDTIME		; no. let's do it then.
OAS1:
	MOV	A, SECONDS	; how many have we seen recently ?
	CJNE	A, #60, TODRET	; return now if it's not a full minute
ONCE_A_MINUTE:			; here once every minute
	MOV	SECONDS, #0	; reset the seconds
	INC	MINUTES		; otherwise it's just like before...
	MOV	A, MINUTES	; ...
	CJNE	A, #60, TODRET	; ...
ONCE_AN_HOUR:			; here once every hour
	MOV	MINUTES, #0	; you must know the story by now...
	INC	HOURS		; ...
	MOV	A, HOURS	; ...
	CJNE	A, #24, TODRET	; ...
ONCE_A_DAY:			; here once every day
	MOV	HOURS, #0	; currently we don't track the date
TODRET:	RET			; all done here


; WAIT,HARDWAIT
;
;   This routine will wait for a fixed period of time.  The time to
; wait is specified in ticks and passed in the A register.  With a
; 120Hz clock, that makes the maximum wait time just over 2 seconds.
; Since this routine works by counting the line time clock, it is
; independent of the processor clock frequency.
;
;   Remember that the clock interrupt routine decrements the WAIT_TIME
; variable "behind our backs". 
; 
; WAIT can be interrupted by a keypress. 
; HARDWAIT CANNOT.
;
;CALL:
;	A/ delay time, in ticks
;	CALL	WAIT		; saves all registers!
;
WAIT:
	MOV	WAIT_TIME, A	; set the timer
WAIT1:  call	getkey
	jz	WAIT2
	ret
WAIT2:
	MOV	A, WAIT_TIME	; then watch it count down
	JNZ	WAIT1		; and wait for it to reach zero
	RET			; that's all there is to it
;
; variant of wait. does not exit prematurely for button press on IR remote
HARDWAIT:
	MOV	WAIT_TIME, A	; set the timer
HARDWAIT1:
	MOV	A, WAIT_TIME	; then watch it count down
	JNZ	HARDWAIT1	; and wait for it to reach zero
	RET			; that's all there is to it


; LONG_BEEP
; SHORT_BEEP
;
;   These two routines will enable the external piezo buzzer and then
; set a time for it to be turned off.  The turn off operation happens
; in the clock interrupt routine and doesn't need anything further
; from us.  Note that this routine does NOT wait for the buzzer to
; finish before returning, so remember that the buzzer may still be
; sounding while your code is executing.  If one of these routines
; are called again before the previous time is finished, then the
; previous timeout will be reset.
;
;CALL:
;	ACALL	LONG_BEEP	; preserves all registers
;	ACALL	SHORT_BEEP	;   ...  ditto  ...
;
LONG_BEEP:
	MOV	BUZZER_TIME, #HERTZ	; a one second beep
	SJMP	BEEP1
SHORT_BEEP:
	MOV	BUZZER_TIME, #HERTZ / 8	; a 1/8 second beep
	sjmp    BEEP1
SHORTSHORT_BEEP:
	MOV	BUZZER_TIME, #HERTZ/40	; a 1/40 second beep
	SJMP	BEEP1
SSSHORT_BEEP:
	MOV	BUZZER_TIME, #HERTZ/60	; a 1/60 second beep
	SJMP	BEEP1
BEEP1:	BUZZER_ON		; and don't forget to turn the thing on!
	RET			; ...


; CLOCK_STARTUP
;
;   This routine is called at system initialization time to start
; up the clock.  It has to enable the clock interrupts (the global
; interrupt enable has already been set) and then "test" the clock.
; This test is part of the power on self test and should check the
; clock for basic function.  What we do here is to simply wait for
; what we know to be a long time and watch for the clock to be ticking.
; The C bit is set on return if the clock is working, and cleared
; if it is not.
;
;CALL:
;	CALL	CLOCK_STARTUP	; uses ALL registers
;	C/1			; iff self test was successful
;
CLOCK_STARTUP:
	SETB	IT1		; make the interrupt edge triggered
	CLR	PX1		; set it to low priority
	SETB	EX1		; and finally enable the interrupt

;   The clock ticks at either 50 or 60Hz; that's either 20 or 16.6
; miliseconds per tick.  This loop takes just over 66,000 cycles to
; execute or, with a 7.3278Mhz clock, just over 108 milliseconds.
; That's enough time to detect at least four clock ticks - if we don't
; get that many, then the clock is busted...
	MOV	R1, #0		; execute the outer loop 256 times
CLKST1:	MOV	R0, #0		; [1] execute the inner loop 256 times
	DJNZ	R0, $		; [2] waste time
	DJNZ	R1, CLKST1	; [2] ...
	MOV	A, TICKS	; see how many ticks we've seen
	CJNE	A, #4, $+2	; have we seen at least 4 ticks?
	JC	CLKST2		; jump if the clock isn't running

; The clock is OK - return success...
	SETB	C		; return C=1 for success
	RET

; The clock isn't running...
CLKST2:	CLR	C		; return C=0 for failure
	RET
