	$TITLE	(MAX7219 Display Controller Software Interface)


; Copyright (C) 1994 by Dan Liddell and Robert Armstrong.
; Copyright (C) 2004 by Spare Time Gizmos.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-
; ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
; Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, visit the website of the Free
; Software Foundation, Inc., www.gnu.org.

;   The display consists of 256 LEDS controlled by four Maxim 7219
; display controller chips.  Each 7219 controls one quadrant of
; the display - upper left (UL), upper right (UR), lower left (LL)
; and lower right (LR).  The 7219 chips contain 64 bits of internal
; RAM for display refresh and are able to scan the display without
; any help from the 8051.
;
;   The display memory in the 7219s is organized as 8 bytes of 8 bits
; each.  Each 7219 contains a total of 14 8 bit registers - 1..8 are
; the actual display data and the remainder are used for control
; functions.  To load a register we must send exactly 16 bits to the
; 7219 - the first (most significant) byte contains the register
; address and the second (least significant) byte contains the register
; data.  Loading a register is the only protocol that the 7219s
; understand - they are pretty simple minded devices.
;
;   Data is sent to the 7219s serially, one bit at a time.  The
; controllers are daisy chained into a serial bus in the order: LR,
; LL, UR, UL.  This means that the first data we send ends up in
; UL and the last data goes to LR.  The serial data and clock signals
; are all generated in software.  Once 64 bits have been clocked out,
; a load signal is activated to cause the 7219s to transfer the data
; into their internal registers.
;
;   The load and clock signals are shared by all 7219s.  There is no
; way to load data into one without loading them all.  Fortunately
; the 7219s do have a no-op register (any data loaded into this register
; is ignored) which we can use to allow loading only one 7219.  In
; the current code this is not necessary.
;
;   Finally, we ought to think briefly about the timing requirements.
; The 7219 have a minimum clock width of 50ns, a minimum load pulse
; width of 50ns, and a data setup time of 25 ns.  Each 7219 in the
; serial bus delays the data (from its serial in to its serial out)
; by 25ns.   At 12Mhz (the fastest we ever expect to use), the 8051
; can execute one instruction per microsecond.  Even if we use adjacent
; instructions to toggle clock and data, we have no fear of any timing
; problems.


; SEND_7219
;
;   This routine will send 8 bits to the 7219 chips and generate the
; required clock signals.  It doesn NOT activate the load signal at
; any time.  Note that the 7219s expect to see the data most significant
; bit first.
;
;   Since we do this a _lot_, we're very concerned about the speed of
; this routine.  It currently takes 59 machine cycles, not counting
; the ACALL that started it.
;
;CALL:
;	A/ data bits to be send
;	CALL	SEND_7219		; uses R2 !!!
;
SEND_7219:
	MOV	R2, #8		; [1] send out a total of 8 bits
SNDSP1:	RLC	A		; [1] shift the next data bit into the carry
	MOV	DISPLAY_DATA, C	; [2] send it to the 7219 data input
	SETB	DISPLAY_CLOCK	; [1] then make the 7219 clock go high
	CLR	DISPLAY_CLOCK	; [1] and low
	DJNZ	R2, SNDSP1	; [2] finish off all 8 bits
	RET			; [2]


; DISPLAY_UPDATE
;
;   This routine will do a complete update of the display controller's
; internal memory from the bitmap in 8051 memory.  This requires loading
; 256 bits into the 7219s, plus the associated register addresses, for
; a total of 512 bits serially shifted out.  The display hardware is
; carefully arranged so that the appropriate bits and bytes for for
; the 7219s appear in the same order as the internal bitmap.  We do have
; to load the top and bottom half of the display at the same time, but
; that's dealt with by keeping two pointers.
;
;   All told, this routine requires just over 4000 machine cycles,
; counting the calls to SEND_7219.
;
;CALL:
;	CALL	DISPLAY_UPDATE	; uses ALL registers !!
;
DISPLAY_UPDATE:
	MOV	R7, #1		; [1] register counter for the 7219 chips
	MOV	R0, #BITMAP	; [1] pointer to the top half of the bitmap
	MOV	R1, #BITMAP+16	; [1]    "    "   "  bottom "  "  "    "

; Send the bits to load the upper left controller...
DSPUP1:	MOV	A, R7		; [1] first send the register address
	ACALL	SEND_7219	; [2]   ... to the 7219s...
	MOV	A, @R0		; [1] then load and send the bitmap data
	ACALL	SEND_7219	; [2] and then load that
	INC	R0		; [1] on to the next bitmap byte

; Now the bits for the upper right controller...
	MOV	A, R7		; [1] send the row address (same as before!)
	ACALL	SEND_7219	; [2]  ... 
	MOV	A, @R0		; [1] then the display data
	ACALL	SEND_7219	; [2]  ...
	INC	R0		; [1] for next time around

; And now the bottom half - lower left first...
	MOV	A, R7		; [1] the row address is the same as the top!
	ACALL	SEND_7219	; [2]  ...
	MOV	A, @R1		; [1] but data from the bottom of the bit map
	ACALL	SEND_7219	; [2]  ...
	INC	R1		; [1] ...

; Lastly the lower right quadrant...
	MOV	A, R7		; [1] you must have the hang of this by now
	ACALL	SEND_7219	; [2]  ...
	MOV	A, @R1		; [1]  ...
	ACALL	SEND_7219	; [2]  ...
	INC	R1		; [1]  ...

; Toggle the LOAD line so the 7219s will latch the current data...
	SETB	DISPLAY_LOAD	; [1] high ...
	CLR	DISPLAY_LOAD	; [1]  ... then low

; That's it - repeat for 8 groups of 4 bytes...
	INC	R7		; [1] count the rows here
	CJNE	R7, #9, DSPUP1	; [2] and repeat as necessary
	RET			; [2] all done


; DISPLAY_COMMAND
;
;   This routine will send a "command" to all of the 7219 chips.
; Sending a command to these devices is just the same as loading
; a register - the difference is in the register address used.
; Special register addresses are used for internal command registers
; in the 7219. 
;
;CALL:
;	R0/	command		; or special register address
;	R1/	data		; argument to the command
;	CALL	DISPLAY_COMMAND	; uses ALL registers !!
;
; NOTE: the same command and data is repeated four times, once for
; each display controller chip.
;
DISPLAY_COMMAND:
	MOV	R7, #4		; repeat once for each 7219
DSPCM1:	MOV	A, R0		; send the command register address first
	ACALL	SEND_7219	; ...
	MOV	A, R1		; then send the data
	ACALL	SEND_7219	; ...
	DJNZ	R7, DSPCM1	; loop for all four chips
	SETB	DISPLAY_LOAD	; then toggle the load line
	CLR	DISPLAY_LOAD	; ... to latch the data in the 7219s
	RET


; SET_INTENSITY
;
;   This routine will set the display intensity to the value specified
; in the A register.  The display intensity is a number from 0 thru F,
; with 0 being the minimum brightness.  The 7219s implement this function
; by changing the display duty cycle.
;
;CALL:
;	A/ brightness		; 0..F
;	CALL	SET_INTENSITY	; uses ALL registers!
;
SET_INTENSITY:
	MOV	INTENSITY, A	; remember the current brightness setting
	MOV	R0, #0AH	; magic address for the intensity register
	MOV	R1, INTENSITY	; and the intensity data
	SJMP	DISPLAY_COMMAND	; load it into the 7219s and return


; DISPLAY_ON
; DISPLAY_OFF
;
;   These routines enable and disable the entire display as a unit.  This
; does not affect the current contents of the 7219 refresh memory, nor
; does it prevent us from sending other commands to the display.  It just
; tells the 7219s to stop scanning and switch off all the digit and
; segment drivers.
;
;CALL:
;	CALL	DISPLAY_ON	; uses ALL registers !
;	CALL	DISPLAY_OFF	; ditto !!!
;
DISPLAY_ON:
	MOV	R1, #1		; set the SHUTDOWN register to 1
	SJMP	DSPOF1		; then fall into the shared code
DISPLAY_OFF:
	MOV	R1, #0		; set the shutdown register to 0
DSPOF1:	MOV	R0, #0CH	; magic address for the shutdown register
	SJMP	DISPLAY_COMMAND	; and ship it out to everybody


; LOAD_BITMAP
;
;   This routine will copy a bitmap from the program memory into the
; bitmap memory.  It's very simple - it just always copies 32 bytes
; of data.  It doesn't actually load the bitmap into the display hard-
; ware - you must call UPDATE_DISPLAY to do that.
;
; CALL:
;	MOV	DPTR, #bitmap
;	CALL	LOAD_BITMAP	; uses R0 and R1
;
LOAD_BITMAP:
	MOV	R0, #BITMAP_SIZE; always copy 32 bytes
	MOV	R1, #BITMAP	; this register contains the index
LDBIT1:	CLR	A		; we have to have an offset
	MOVC	A, @A + DPTR	;  to access program memory
	MOV	@R1, A		; copy it to the bitmap
	INC	R1		; on to the next byte
	INC	DPTR		; ...
	DJNZ	R0, LDBIT1	; do them all
	RET
	

; DISPLAY_STARTUP
;
;   This routine is called at system initialization time and it has a 
; dual purpose.  The first is to perform a power on self test of the
; display hardware, and the second is to initialize the display and
; leave it in a state for displaying graphics.
;
;   The self test is implemented via a hardware assist.  The serial
; data out of the last 7219 in the chain is connected to an 8051
; I/O port bit.   What we do in the self test is to clock out known
; data, and then watch for it to come back to us 64 bit times later.
; While this isn't the most complete test, it does verify that all
; four 7219 chips are present and talking.
;
;CALL:
;	CALL	DISPLAY_STARTUP	; uses ALL registers
;	C/1			; iff self test was successful
;
DISPLAY_STARTUP:
	CLR	DISPLAY_DATA	; set the initial state of all the
	CLR	DISPLAY_CLOCK	;  MAX7219 control signals...
	CLR	DISPLAY_LOAD	; ...
	SETB	DISPLAY_ECHO	; ...

;   At this time, the 7219 serial registers contain unknown data.  First
; we clock out 64 zero bits to get them all into a known state.
	CLR	DISPLAY_DATA	; always send zeros
DSPST0:	MOV	R0, #64		; number of bits to send
DSPST1:	SETB	DISPLAY_CLOCK	; toggle the clock
	CLR	DISPLAY_CLOCK	; ...
	DJNZ	R0, DSPST1	; ...

;   Now pump out 64 one bits.  At the same time, monitor the data shifted
; out of the last 7219 to ensure that all bits are zeros!
	MOV	R0, #64		; same number of bits as before
	SETB	DISPLAY_DATA	; but send ones this time
DSPST2:	MOV	C, DISPLAY_ECHO	; read the data output from the last chip
	JC	DSPST9		; if it's one the display isn't working
	SETB	DISPLAY_CLOCK	; toggle the clock
	CLR	DISPLAY_CLOCK	;  (data is clocked on the falling edge!)
	DJNZ	R0, DSPST2	; do all the bits

;   One last time - clock out 64 zeros (to set the display data registers
; to a noop state) and watch for 64 one bits coming back...
	MOV	R0, #64		; you've seen all this before
	CLR	DISPLAY_DATA	; ...
DSPST3:	MOV	C, DISPLAY_ECHO	; ...
	JNC	DSPST9		; this time we're looking for ones!
	SETB	DISPLAY_CLOCK	; ...
	CLR	DISPLAY_CLOCK	; ...
	DJNZ	R0, DSPST3	; ...

;   Just for grins, use the 7219 display test function to turn on all
; LEDs.  Display test overrides all other controller settings, so this
; will work even though we haven't initialized the display yet.  After
; half a second, return the display to normal mode.  Of course, we can't
; from in here whether the LEDs actually came on, but there might be
; somebody else out there watching.
	MOV	R1, #1		; select display test mode
	MOV	R0, #0FH	; test register address
	ACALL	DISPLAY_COMMAND	; ...
	MOV	A, #HERTZ	; wait for 1 second
	lCALL	HARDWAIT	; *****
	MOV	R1, #0		; return to normal mode
	ACALL	DISPLAY_COMMAND	; ...

;   Ok, enough testing.  Initialize the internal display controller
; registers to the values appropriate to our hardware.
	MOV	R1, #7		; enable all eight digit (column) drivers
	MOV	R0, #0BH	; address for the scan limit register
	ACALL	DISPLAY_COMMAND	; tell 'em all!
	MOV	R1, #0		; set no decode mode for all digits
	MOV	R0, #09H	; decoder mode register address
	ACALL	DISPLAY_COMMAND	; ...
	MOV	A, #0FH		; set the intensity to maximum
	ACALL	SET_INTENSITY	; ...

;   Load the internal 7219 refresh data from our internal bitmap.  Note
; that the bitmap has been cleared to all zeros during RAM initialization!
	ACALL	DISPLAY_UPDATE	; away we go!

; That's it - turn on the display and return success...
	ACALL	DISPLAY_ON	; clear the shutdown mode
	SETB	C		; return a successful status
	RET			; ...

; We get here if the display fails the self test for any reason...
DSPST9:	CLR	C		; return a failure status
	RET			; ...