; Copyright (C) 1994 by Dan Liddell and Robert Armstrong.
; Copyright (C) 2004 by Spare Time Gizmos.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-
; ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
; Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, visit the website of the Free
; Software Foundation, Inc., www.gnu.org.

; WARNING!!!
;
;   The code for both the IR and Timer 0 interrupt use register bank #1.
; They assume that they don't need to save any of these registers, and
; they further assume that whatever they left in those registers at the
; last interrupt will still be there this time.  In other words, don't
; touch!
;
;   Register usage in both IR_INTERRUPT and TIMER0_INTERRUPT:
;
;	R0, R1	-> general purpose temporaries
;	R2	-> flags
;		R2.0 -> this is a bad code if this bit is set
;	R3	-> (unused)
;	R4	-> the last two samples made
;	R5	-> counts the number of samples
;	R6:R7	-> RC5 codeword being assembled (R6 is the MSB)

	USING	1

; Masks for flag bits stored in R2...
RC5ERR	EQU	80H	; the current transmission is invalid

; Select the RC5 register set...
RC5REGS	MACRO
	CLR	RS1
	SETB	RS0
	ENDM


; GENERAL
;
;   RC5 codes transmit 14 bits using biphase encoding with a nominal
; bit time of 1.778 ms.  In biphase, each data bit is transmitted as
; two signalling units.  For a zero data bit, a 1/2 bit time of an IR
; carrier followed by a 1/2 bit time of no carrier is transmitted.  For
; a one data bit it's a half bit time of no carrier followed by a half
; bit of carrier.  The key feature is that there's always a transition
; in the middle of a bit.  Sometimes there's a transition at the beginning
; of a bit, but not always.
;
;   To recover the original data, we arrange for timer 0 to interrupt
; every 1/2 bit time in the middle of each signalling unit.  We sample
; the IR data at each interrupt, verify that we see either no-carrier/
; carrier or carrier/no-carrier pairs, and recover the original bit.
; Fourteen bits are reassembled in a software shift register (the MSB
; is always transmitted first!) to recreate the original data.
;
;   This would work fine, except that there is quite a bit of variation
; in the bit timing of the IR remotes.  Anywhere from 3 to 5 percent
; is not unusual.  This doesn't sound like much, but the error accumulates
; for each bit.  By the fourteenth bit, our sample window can easily be
; half a bit time in error.  Naturally that doesn't work very well.
;
;   To get around this, we leave the IR INTERRUPT (INT 0) enabled.
; Everytime there is a positive edge (a transition from no-carrier to a
; carrier) on the IR data we get an interrupt.  We use this interrupt
; to reset the timer.  Since the worst case time between positive edges
; is two bit times, the biggest timing error we can expect is 6 to 10%.
; That we can live with.


; INTERFACE
;
;   This code is pretty much self maintaining once IR_STARTUP is called and
; the interrupt system turned on.  All the important work of receiving
; and assembling IR codes happens at interrupt level.  Once a complete
; RC5 code has been received, we will store the entire 14 bit code in
; the RAM locations IRDATA (MSB) and IRDATA+1 (LSB).
;
;   The routine GETKEY will examine IRDATA for a non-zero value and then
; translates in into an internal key code.  This translation is completely
; table driven - see the table located at RC5KEYS.
;
;   The storage at IRDATA serves as a simple one word buffer for IR key
; strokes. There is no more, and there is no detection for overflow.  If
; a new RC5 code is received before the previous one has been read, then
; the old one is simply lost.  FYI, a RC5 remote will repeat keys just
; slightly slower than 10Hz.  That puts the lower bound on how often the
; background must poll to avoid losing data.



; IR_INTERRUPT
;
;   This routine is attached to INT 0, the IR data bit.  It has two
; major functions - while we are busy receiving an RC5 code it must
; resynchronize our timer with the RC5 bit stream.  While we are not
; receiving anything, this routine gets the first indication that
; something is coming in.  In that case it has to get everything started
; up for receiving the code.
;
;    WARNING - be sure to adjust the timing constants in both IRSTART and
; IRSYNC if you change the following instructions!
;
IR_INTERRUPT:
	PUSH	ACC		; [2] we need to save this much room to work
	PUSH	PSW		; [2] and the current status
	RC5REGS			; [2] select register bank 1
	CJNE	R5, #0, IRSYNC	; [2] jump if we are receiving something

;   If we get here, then we've jsut begun to receive a transmission.  The
; very first thing that we want to do is to start up the timer.  We do this
; right away so that we won't have to worry the timing errors any delays
; would cause.  Since this ISR and the Timer 0 ISR share the same interrupt
; priority, we don't have to worry about the timer going off before we
; finish initialization.
;
;   Since the first bit of every RC5 code is a 1, when we get here we
; know we're in the _middle_ of a bit.  Since we want to sample in
; the middle of each half-bit signalling unit (i.e. at 1/4 and 3/4 of
; a bit time), we initialize the timer to go off in 1/4 of a bit.
;
;  One last thing - the 8051 took an average of 4 cycles to start this
; ISR.  Our code took 8 cycles getting here.  The Timer 0 ISR requires
; another 4 plus 5 cycles to make a sample.  It'll take the following
; code another 6 cycles to get the thing started.  This totals to 26
; cycles, which is subtracted from the timer setting.
IRSTART:
	MOV	TH0, #HIGH(-(RC5QBIT-26))	; [2]
	MOV	TL0, #LOW( -(RC5QBIT-26))	; [2]
	SETB	TR0		; [1] NOW! we're timing from here on!

; Now we can do the rest of the initialization at our leisure...
	MOV	R6, #0		; clear the code accumulator
	MOV	R7, #0		; ...
	MOV	R5, #1		; we've already made one sample
	MOV	R4, #0		; ... and that sample was a zero!
	MOV	R2, #0		; clear the flags register
	SJMP	IREXIT		; restore the context and return

;   We get here for an IR interrupt while we are in the middle of reading
; a code.  This interrupt happens on any positive edge in the IR data,
; and it could be either the beginning of a bit or the middle of a bit.
; With a little thought we could figure out which it is, but we really
; don't care.  Since we always sample twice per bit, at 1/4 and 3/4 of
; the way thru, we know that it's always 1/4 of a bit time to the next
; sample.
;
;   The time correction here is 27 cycles.  See the code at IRSTART for
; a discussion of this - the only difference is that here it takes an
; extra cycle to get going.
IRSYNC:	CLR	TR0		; [1] stop the timer from running
	MOV	TH0, #HIGH(-(RC5QBIT-27))	; [2]
	MOV	TL0, #LOW( -(RC5QBIT-27))	; [2]
	SETB	TR0		; [1] start running again
IREXIT:	POP	PSW		; restore original context and return
	POP	ACC		; ...
	RETI			; ...



; TIMER0_INTERRUPT
;
;   This routine is called at regular intervals (1/2 a bit time, to be
; exact) while we are receiving.  It samples the IR data, resets the
; timer, and then figures out the data received.  Watch closely and
; hold on tight.
;
;   WARNING!  If you change any of the initial instructions of this
; ISR, you'll have to recalculate the timer corrections here, and at
; IRSTART and IRSYNC.
;
TIMER0_INTERRUPT:
	PUSH	ACC		; [2] save some working room
	PUSH	PSW		; [2] ...
	MOV	C, IRBIT	; [1] OK! Data is sampled here !!

;   Now restart the timer.  Don't let anything happen to the carry bit!!
; The timer correction here is 4 cycles for the ISR latency, 5 cycles for
; the code above, and another 6 to get it restarted.  Total is 15 cycles.
	CLR	TR0		; [1] stop the timer
	CLR	TF0		; [1] clear the flag
	MOV	TH0, #HIGH(-(RC5HBIT-15))	; [2]
	MOV	TL0, #LOW (-(RC5HBIT-15))	; [2]
	SETB	TR0		; [1] and start running again

;   Now process the data we sampled.  Remember that we interrupt twice for
; each bit.  The first time, we just save the IR state away for the next
; interrupt.  For the second interrupt, we verify that the current IR state
; is the complement of the previous state and then store the bit.  Remember
; that the current IR sample is still in the C bit!!
	RC5REGS			; select register bank 1
	MOV	A, R4		; add the current sample to R4
	CPL	C		; (remember that the IR data is active low!)
	RLC	A		; ...
	MOV	R4, A		; ...
	INC	R5		; count another sample made
	MOV	A, R5		; and get the sample number
	JB	ACC.0, IREXIT	; we can quit now if this is an odd sample

;   We've accumulated two samples.  Make sure they represent a valid
; biphase code and figure out what the data bit must have been.  Add
; that data bit to the RC5 accumulator in R6:R7.
	CJNE	R4, #1, TMRIN1	; jump if the data bit wasn't a 1
	SETB	C		; remember the data bit
	SJMP	TMRIN2		; ...
TMRIN1:	CJNE	R4, #2, TMRIN3	; jump if the data bit wasn't a 0
	CLR	C		; put the data bit in C
TMRIN2:	MOV	A, R7		; get the low order RC5 code
	RLC	A		; shift in the new data bit
	MOV	R7, A		; ...
	MOV	A, R6		; do a 16 bit double precision shift
	RLC	A		; ...
	MOV	R6, A		; ...
TMRIN4:	MOV	R4, #0		; clear the sample register
	CJNE	R5, #28, IREXIT	; if this isn't the last bit, then return
	SJMP	IRDONE		; go do the post reception processing

;   Here if the two samples don't make a valid biphase code.  We're sort of
; stuck.  There's no way we can abort the rest of the transmission, so we
; just have to wait it out. We can set a flag in R2 to indicate that this
; transmission is garbage and when we finally get to the end we will just
; ignore the whole transmission.
TMRIN3:	MOV	R2, #01H	; set the error flag in R2
	SJMP	TMRIN4		; and then just dismiss the interrupt

;   We've received all 14 bits now.  First, turn off the timer and disable
; it!  Then reset the bit counter in R5 so that the next IR interrupt will
; start a brand new code.  If there was an error in the code, then just
; discard it without storing anything.  Otherwise store both bytes
; in the RAM locations IRDATA.
IRDONE:	CLR	TR0		; stop the timer
	MOV	R5, #0		; reset for the next RC5 code
	MOV	A, R2		; get the reception flags
	JB	ACC.0, IREXIT	; don't store anything if it was bad
	MOV	IRDATA, R6	; store the MSB
	MOV	IRDATA+1, R7	; ... and the LSB of the RC5 data

;	mov bitmap+14,r6	; debug
;	mov bitmap+15,r7	; debug. what are we getting here?
;	call shortshort_beep	; debug. are we getting any damn keys?
	
	SJMP	IREXIT		; and then return

; IR_STARTUP
;
;   This routine is called at system initialization time to startup the
; IR receiver.  It will perform all the necessary hardware and software
; initialization for the IR interrupt (INT 0) and the timer 0 interrupt.
; It assumes that the global interrupt enable has already been set.
; Normally this routine is also supposed to perform a self test of the
; hardware ane return a pass/fail status, but for the IR port we don't
; really have any self test.
;
;   Be warned that there just could be some bozo out there with his
; finger planted on the IR remote right now, so be sure that everything
; is ready before enabling the IR interrupts!
;
;CALL:
;	CALL	IR_STARTUP	; uses ALL registers
;	C/1			; always gives a success return
;
IR_STARTUP:

;   First initialize the IR receiver data.  Remember that it uses
; register bank #1.  The only register that really needs to be
; initialize is R5 - the rest will be set appropriatly when the first
; IR bit comes in.
	PUSH	PSW		; save the current register bank
	RC5REGS			; select register bank 1
	MOV	R5, #0		; reset the receiver sample counter
	MOV	IRDATA, #0	; reset the current code
	MOV	IRLAST, #0	; ...

;   The intervals we want to time are just too large for 8 bits, so we
; have to use the timer in 16 bit mode.  Setup timer 0 for our use but
; don't actually start it now - it'll be started on the first IR
; interrupt.  Don't bother the settings of timer 1 !!
	ANL	TMOD, #0F0H	; clear timer 0 settings
	ORL	TMOD,  #01H	; mode 1, interval timer
	CLR	TR0		; make sure it isn't running
	CLR	TF0		; and be sure the flag is cleared

;   External interrupt 0 is the IR data.  Although the actuall interrupt
; pin on the 8051 is active low, the Sharp IR receiver also inverts the
; IR data.  So from the software perspective INT 0 is positive edge
; triggered - we get an interrupt when the IR goes from a "no-carrier"
; state to a "carrier detected" state.
	SETB	IT0		; make the interrupt edge triggered

;   Because the timing of the IR interrupts is so critical, we make
; our two interrupts the only high priority ones.  Every other
; interrupt is set to low priority.
	MOV	IP, #0		; make everything low priority
	SETB	PX0		; make IR data high
	SETB	PT0		; and Timer 0 high priority

;   Ok, here we are.  Enable timer 0 interrupts, then IR data interrupts,
; then hold your breath...
	SETB	ET0		; enable Timer 0
	SETB	EX0		; and INT 0

; All done...
	POP	PSW		; restore the original register bank
	SETB	C		; pretend the self test passed
	RET			; and we're outta here


	USING	0		; GETKEY uses the normal register set!

; GETKEY
;
;   This routine will examine the IRDATA bytes for a non-zero value.
; If it finds one, then it will try to translate the code using the
; remote control key table.  The resulting key code, a single byte,
; is returned in A.  If IRDATA is zero, then A will be cleared upon
; return.  After reading the IRDATA it will be cleared for the next
; time.  This means that you can read each keystroke only once!
;
;   A word on autorepeat for remotes: the remote will immediately
; start repeating the transmission as long as the user holds the
; button down.  In most cases we want to ignore this repetition and
; register just one key code.  We do this quite simply by remembering
; the last key code we received and discarding this one if it is the
; same.  RC5 remotes have the nice feature that they will toggle the
; control bit in the code each time the button is pressed.  This means
; that we be able to differentiate two presses of the same button
; from autorepeat.
;
;   Certain keys, as indicated in the key table, allow repetition.
; When we receive one of these keys we simply clear the memory of
; the last key code.  This guarantees that the next key will not
; match even if it happens to be exactly the same.
;
;CALL:
;	CALL	GETKEY		; uses all registers _and_ DPTR!
;	<keycode in A>		; or zero if no key detected
;
GETKEY:

;   See if there's any IRDATA there at all.  If there is, then turn off
; interrupts and copy it out of the table.  This ensures that the
; interrupt routines won't change the IRDATA while we're thinking!  If
; this IR code is the same as the last one received, then ignore it.
	MOV	A, IRDATA	; see if there's anything there
	JZ	GETKE1		; return zero if there's no key waiting
	CLR	EA		; interrupts off!
	MOV	R6, IRDATA	; make a copy of the current code
	MOV	R7, IRDATA+1	; ...
	MOV	IRDATA, #0	; and clear the code for next time
	SETB	EA		; interrupts back on!
	MOV	A, R6		; see if this code is the same as last time
	CJNE	A,IRLAST,GETKE2	; jump if it's not
	MOV	A, R7		; we have to test both bytes
	CJNE	A,IRLAST+1,GETKE2; ...

; Here if there's no key for one reason or another.  Just return zero.
GETKE1:	
	CLR	A		; say that there's no key
	RET			; ...

;   We've got a valid code.  Remember it as the "last key" for next
; time and prepare to search the key code table...
GETKE2:	MOV	IRLAST, R6	; remember this keycode for next time
	MOV	IRLAST+1, R7	; ...
	MOV	DPTR, #RC5KEYS	; load the address of the table
	ANL	AR6, #0F7H	; ignore the RC5 control bit now
	MOV	R1, #0		; keep the table index here

;   Now search the RC5KEYS table for a match.  If we don't find one,
; then we assume that this is a bad code and ignore it.  This is just
; a simple linear search, but we'll assume that there are never going
; to be too many keys in the table!
GETKE3:	MOV	A, R1		; get the table index
	INC	R1		; then increment it
	MOVC	A, @A+DPTR	; get a byte from the table
	JZ	GETKE1		; jump if this is the end of the table.
	CJNE	A, AR6, GETKE4	; jump if this entry doesn't match
	MOV	A, R1		; check the second byte next
	MOVC	A, @A+DPTR	; now try the next byte
	CJNE	A, AR7, GETKE4	; jump if it doesn't match
	SJMP	GETKE5		; we've found a match!!
GETKE4:	INC	R1		; on to the next table entry
	INC	R1		; ...
	INC	R1		; ...
	SJMP	GETKE3		; then go try again

;   Here if we actually found a match.  Check out the flag byte from the
; table first.  Currently the only flag bit defined is bit 0, which says
; that this key is allowed to repeat.  We implement that by simply
; forgetting that we saw this key.
GETKE5:	INC	R1		; advance to the flag byte
	MOV	A, R1		; then get it out of the table
	MOVC	A, @A+DPTR	; ...
	JNB	ACC.0, GETKE6	; skip if the flag bit isn't set
	MOV	IRLAST, #0	; allow this key to repeat immediately

; Now get the key code from the table and load it into A.
GETKE6:	INC	R1		; point to the last byte
	MOV	A, R1		; ...
	MOVC	A, @A+DPTR	; get it out of the table
	call shortshort_beep	; good feedback for key presses. not debug
	RET			; and we're outta here...



; RC5 Code Table
;
;   This table is used to translate RC5 codes into internal "key"
; codes that are returned by GETKEY.  Each entry in the table
; consists of four bytes:
;
;	bytes 0:1 -> the RC5 code, MSB first
;	byte 2    -> flag byte
;	byte 3    -> internal key code for GETKEY to return
;
;   The only flag currently defined is bit 0, which indicates that this key
; can  be repeated.  The key code can be any non-zero value.  The RC5
; code should always have the control bit cleared.
;
;   This four byte entry is repeated once for each key on the remote.
; The end of the table is marked by a single zero byte.
;
RC5KEYS:

; Dan's original key bindings for the Magnavox 8 key remote...
	DB	34H, 7FH, 0, CMD_COPYRIGHT	;		    "TUNER"
	DB	35H, 3FH, 0, CMD_NEXTMODE	;		    "CD"
	DB	35H, 20H, 0, CMD_BRIGHTNESS	;		    "PRESET"
	DB	34H, 60H, 0, CMD_BRIGHTNESS	;		    "PRESET"
	DB	35H, 0CH, 0, CMD_SHOWMENU 	; (CMD_ZEROSECONDS) "STANDBY"
	DB	34H, 4CH, 0, CMD_SHOWMENU 	; (CMD_ZEROSECONDS) "STANDBY"
	DB	35H, 36H, 0, CMD_AUTOADVANCE 	; (CMD_STOPCLOCK)   "STOP"
	DB	34H, 76H, 0, CMD_AUTOADVANCE 	; (CMD_STOPCLOCK)   "STOP"
	DB	35H, 35H, 1, CMD_NEXTPATTERN	; (CMD_SETHOURS)    "PLAY"
	DB	34H, 75H, 1, CMD_NEXTPATTERN	; (CMD_SETHOURS)    "PLAY"
	DB	35H, 11H, 1, CMD_SLOWER		; (CMD_SETSECONDS)  "VOL DN"
	DB	34H, 51H, 1, CMD_SLOWER		; (CMD_SETSECONDS)  "VOL DN"
	DB	35H, 10H, 1, CMD_FASTER		; (CMD_SETMINUTES)  "VOL UP"
	DB	34H, 50H, 1, CMD_FASTER		; (CMD_SETMINUTES)  "VOL UP"

; New key bindings for universal remote programmed for a Philips TV...
	DB	30H, 22H, 0, CMD_COPYRIGHT	;		    "PREV"
	DB	30H, 0DH, 0, CMD_NEXTMODE	;		    "MUTE"
	DB	30H, 0CH, 0, CMD_BRIGHTNESS	;		    "POWER"
	DB	30H, 2EH, 0, CMD_SHOWMENU 	; (CMD_ZEROSECONDS) "MENU"
	DB	30H, 26H, 0, CMD_AUTOADVANCE 	; (CMD_STOPCLOCK)   "SLEEP"
	DB	30H, 10H, 1, CMD_NEXTPATTERN	; (CMD_SETHOURS)    "VOL +"
	DB	30H, 20H, 1, CMD_SLOWER		; (CMD_SETSECONDS)  "CH UP"
	DB	30H, 21H, 1, CMD_FASTER		; (CMD_SETMINUTES)  "CH DN"

; End of table...
	DB	0
